# ---------------------------------------------------------------------------- #
#                                Swap Variables                                #
# ---------------------------------------------------------------------------- #

# %%

x = 1
y = 2

x = y
y = x

tmp = x
x = y
y = tmp

x, y = y, x

#%%

# ---------------------------------------------------------------------------- #
#                                 Conditionals                                 #
# ---------------------------------------------------------------------------- #

x, y = 1, 2
if x < y:
    print(y)
elif x > y:
    print(x)
else:
    print(x, y)

# %%

# ---------------------------------------------------------------------------- #
#                                     Loops                                    #
# ---------------------------------------------------------------------------- #

T = (0.1, 0.1)
x = 0
for i in range(len(T)):
    for j in T:
        x += i + j
        print(x)
print(i)

# %%

# ---------------------------------------------------------------------------- #
#                                Guess and Check                               #
# ---------------------------------------------------------------------------- #

# checking if a number is a perfect square
"""
    NOTE: This is not the best to approach this. What could be a better way
    to approach this problem?
"""
number = int(input("Enter a number: "))
answer = 0
while answer**2 < abs(number):
    answer += 1
if answer**2 == abs(number):
    print(f'Square root of {number} is {answer}.')
else:
    print(f'{number} is not a perfect square')

# %%

# ---------------------------------------------------------------------------- #
#                                     Scope                                    #
# ---------------------------------------------------------------------------- #

def scope_example(x, y):
    x, y = y, x
    print(x)
x = 3
y = -3
print(x)
scope_example(x, y)
print(x)

# %%

def list_scope(L1, L2):
    L1 = [9, 8, 7, 6]
    L2.append(9)
    print(f'In list_scope, L1 = {L1}, L2 = {L2}')
L1 = [1, 2, 3, 4]
L2 = [5, 6, 7, 8]
list_scope(L1, L2)
print(f"Outside list_scope, L1 = {L1}")
print(f"Outside list_scope, L2 = {L2}")

# %%

# Recursion

def deepcopy(L):
    """
    Args:
        L: a potentially nested list of lists
    
    Returns:
        list: An exact copy of L (with nested structure)
    """

    ret = []

    for i in L:
        if type(i) is int:
            # base case
            ret.append(i)
        else:
            # recursive case
            ret.append(deepcopy(i))
    
    return ret

original = [1, [2, [3, [4]]]]
copy = deepcopy(original)

copy[1][1][1][0] = "MODIFIED"
print(original)
print(copy)

# %%

def cartesian_product(L):
    """
    Args:
        L: a list of lists

    Returns:
        list: A list of tuples where each tuple is created by selecting one element from each sublist of L.
    """
    # base case
    if len(L) == 0:
        return []
    
    if len(L) == 1:
        options = L[-1] # equivalent to L[0]
                        # used L[-1] for consistency with recursive case
        return [(i,) for i in options]

    # recursive case
    existing_combos = cartesian_product(L[:-1])

    ret = []
    for combo in existing_combos:
        options = L[-1]
        for i in options:
            ret.append(combo + (i,))
        
    return ret

print(cartesian_product([]))
print(cartesian_product([["leather boots", "white sneakers", "black sneakers"]]))
print(cartesian_product([["red shoes", "white shoes", "black shoes"], ["blue jeans", "gray jeans", "tan khakis"]]))


# %%

def permutations(L):
    """
    Args:
        L (list): A list of unique ints to generate the permutations of.
    
    Returns:
        list: A list of tuples, where each tuple is a permutation of L.
    """

    # base cases
    if len(L) == 0:
        return []
    
    if len(L) == 1:
        return [(L[0],)]
    
    # recursive case
    i = L[-1]
    permutations_without_i = permutations(L[:-1])
    
    permutations_with_i = []

    for perm in permutations_without_i:
        for insert_pos in range(len(perm)):
            permutations_with_i.append(
                perm[:insert_pos] + (i,) + perm[insert_pos:]
            )
        # the for loop isn't able to handle adding i to the end (would give IndexError)
        # so we have to do that ourselves
        permutations_with_i.append(perm + (i,))

    return permutations_with_i

print(permutations([]))
print(permutations([1]))
print(permutations([1, 2]))
print(permutations([1, 2, 3]))
print(permutations([1, 2, 3, 4]))

# %%

# Flatten nested lists
def flatten(L):
    """Flattens a nested list of ints.

    Args:
        L (list): A nested list of ints
    
    Returns:
        list: A flattened list of ints.
    """
    ret = []

    for i in L:
        if type(i) is not list:
            # base case
            ret.append(i)
        else:
            # recursive case
            ret.extend(flatten(i))
    
    return ret

print(flatten([1, [2, [3]]]))
print(flatten([[1], 2, 3, [[4, [[5, 6, 7], 8, 9]], 10]]))

# %%

# Classes
class Vehicle(object):
    def __init__(self, name):
        # an instance attribute
        self.name = name

    def get_name(self):
        return self.name

    def honk(self):
        # an instance method
        print(f"{self.name} says HONK")

my_vehicle = Vehicle("batmobile")
print(my_vehicle.get_name())   # prints: batmobile
my_vehicle.honk()              # prints: batmobile says HONK


# %% 
class Car(Vehicle):
    def __init__(self, name):
        super().__init__(name)
        self.type = "car"

    def honk(self):
        print(f"{self.name} says BEEP")

class Truck(Vehicle):
    def __init__(self, name):
        super().__init__(name)
        self.type = "truck"

    def honk(self):
        print(f"{self.name} says BIGHONK")

# Polymorphism
vehicles = [Truck("matt"), Vehicle("ashhad"), Car("esha")]

for v in vehicles:
    v.honk()
# %%
